#include<iostream.h>
#include"reliable.h"
#include<assert.h>
#include<unistd.h>
#include<errno.h>
#include<stdlib.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<sys/mman.h>
#include"debug.h"

static EnvDebug debug("ReliableSegment","ReliableSegment");
#define DEBUGOUT cerr

int TwinblockStrategy::open(char *filename)
{
  f=::open(filename,O_RDWR);
  if(f>0) return 0;
  if(errno == ENOENT)
  {
    f=::open(filename,O_RDWR|O_CREAT,0600);
    if(f>0)
    {
      return 1;
    }
    else
    {
      cerr << "Can't open " << filename << endl;
      exit(1);
    }
  }
  return -1;
}
void TwinblockStrategy::close() { ::close(f); }
int TwinblockStrategy::nplex() { return 2; }
int TwinblockStrategy::fd(int) { return f; }
off_t TwinblockStrategy::offset(int i,int pageno)
{
  return (nplex()*pageno+i)*BLOCKSIZE;
}

uint32_t ReliableSegment::calc_checksum(octet_t *page)
{
  uint32_t* i;
  uint32_t ret = 0;
  for(i = (uint32_t *)page;i < (uint32_t *)(page+RELPAGESZ);i++)
  {
    ret ^= *i;
  }
  return ret;
}

ReliableSegment::ReliableSegment(ReliableSegStrategy *s): PMSeg()
{
  unsigned int lang=maxblocks*RELPAGESZ;
  strat=s;
  fd0=::open("/dev/zero",O_RDONLY);
  if(fd0 < 0) { cerr << "open(/dev/zero) failed" << endl; exit(1); }
  char *puffer=(char *)mmap(NULL,lang,PROT_READ,MAP_PRIVATE,fd0,0);
  if(puffer==(char *)MAP_FAILED)
  {
    cerr << "mmap(/dev/zero) failed" << endl; exit(1);
  }
  set(puffer,lang);
  for(int p=0;p<maxblocks;p++) vdata[p].dirty=false;
}

void ReliableSegment::write_page(int pageno)
{
  struct iovec iv[2];
  VerwData_t *vd = &(vdata[pageno].pverw);
  octet_t *p = page(pageno);
  vd->checksum = calc_checksum(p);
  time(&(vd->stamp));

  iv[0].iov_base=vd;iv[0].iov_len=sizeof(VerwData_t);
  iv[1].iov_base=p; iv[1].iov_len=RELPAGESZ;
  for(int i=0;i< strat->nplex();i++)
  {
    int r=lseek(strat->fd(i),strat->offset(i,pageno),SEEK_SET);
    if(r<0) continue;
    writev(strat->fd(i),iv,2);
    fsync(strat->fd(i));
  }
  vdata[pageno].dirty=false;
  protect(pageno);
}

void ReliableSegment::read_page_optimistic(int pageno)
{
  struct iovec iv[2];
  VerwData_t *vd = &(vdata[pageno].pverw);
  octet_t *p = page(pageno);
  unprotect(pageno);
  iv[0].iov_base=vd;iv[0].iov_len=sizeof(VerwData_t);
  iv[1].iov_base=p; iv[1].iov_len=RELPAGESZ;
  
  int ok=(lseek(strat->fd(0),strat->offset(0,pageno),SEEK_SET)>=0);
  ok = ok && ((size_t)readv(strat->fd(0),iv,2)==BLOCKSIZE);
  ok = ok && (calc_checksum(p) == vd->checksum);
  if(!ok)
  {
    read_page_reliable(pageno);
  }
  vdata[pageno].dirty=false;
  protect(pageno);
}

void ReliableSegment::read_page_reliable(int pageno)
{
  static char lpuffer[BLOCKSIZE];
  char *data = lpuffer+sizeof(VerwData_t);
  VerwData_t *vd = (VerwData_t *)lpuffer;
  unprotect(pageno);
  bool atleastone=false;
  bool badblockfound=false;
  for(int i=0;i<strat->nplex();i++)
  {
    int ok=(lseek(strat->fd(i),strat->offset(i,pageno),SEEK_SET)>=0);
    ok = ok && ((size_t)read(strat->fd(i),lpuffer,BLOCKSIZE)==BLOCKSIZE);
    ok = ok && (calc_checksum((octet_t*)data) == vd->checksum);
    if(!ok) { badblockfound=true; continue; }
    bool newer=false;
    if(!atleastone || (newer=(vd->stamp > vdata[pageno].pverw.stamp)) )
    {
      memcpy((char *)&(vdata[pageno].pverw),vd,sizeof(VerwData_t));
      memcpy((char *)begin()+pageno*RELPAGESZ,data,RELPAGESZ);
      badblockfound = badblockfound || newer;
      atleastone=true;
    }
  }
  if(!atleastone) { cerr << "Fatal: can't read any block" << endl; exit(1); }
  if(badblockfound) write_page(pageno);
  vdata[pageno].dirty=false;
  protect(pageno);
}


void ReliableSegment::protect(int pageno)
{
  int r=mprotect(page(pageno),RELPAGESZ,PROT_READ);
  if(r<0) 
  {
    cerr << "ReliableSegment::protect: mprotect: " << strerror(errno) << endl;
    exit(1);
  }
}

void ReliableSegment::unprotect(int pageno)
{
  int r=mprotect(page(pageno),RELPAGESZ,PROT_WRITE|PROT_READ);
  if(r<0) 
  {
    cerr << "ReliableSegment::protect: mprotect: " << strerror(errno) << endl;
    exit(1);
  }
}

void  ReliableSegment::handle(void *page,void *addr)
{ 
  int pageno= ((char *)page - (char *)begin()) >> 12; // PAGESIZE=2^12
  vdata[pageno].dirty=true;
  if(debug.level(2))
  {
    DEBUGOUT << "Page " << pageno << " is becomming dirty ... " << endl;
  }
}

void ReliableSegment::save()
{
  debug.message(1,"ReliableSegment::save()\n");
  for(int p=0;p<maxblocks;p++)
  {
    if(vdata[p].dirty)
    {
      if(debug.level(2)) DEBUGOUT << "\tSaving page " << p << " ... " << endl;
      write_page(p);
    }
  }
}
#ifdef RELIABLE_TEST
int main(int argc,char **argv)
{
  
  if(argc < 3) { cerr << "Aufruf: -r[r]|-w dateiname" << endl; exit(1); }
  
  TwinblockStrategy tws;
  ReliableSegment rseg(&tws);
  int r=rseg.open(argv[2]);
  if(strcmp(argv[1],"-w")==0)
  {
    int i=0;
    for(char *p = (char *)(rseg.begin());p< (char *)(rseg.end());p++)
    {
      *p= (char)('A'+ i%26);
      i++;
    }
    rseg.write_page(0);
  }
  else
  {
    if(r==1) 
    {
      cerr << "Call with -w first ! "
	   << "(In applications the initialization could be done here )"
	   << endl;
      exit(-1);
    }
    if(strcmp(argv[1],"-rr")==0)
    {
      rseg.read_page_reliable(0);
    }
    else
    {
      rseg.read_page_optimistic(0);
    }

    for(char *p = (char *)(rseg.begin());p< (char *)(rseg.begin())+20;p++)
    {
      cout << *p;
    }
    cout << " ... " << endl;

    cerr << "Changing byte 5 ..." << endl;
    char *ptr =(char *)(rseg.begin())+5;
    *ptr='X';
  }
  rseg.save();
  rseg.close();
}
#endif





