#include<sys/mman.h>
#include<fcntl.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/errno.h>
#include<stdlib.h>
#include"shadowsegment.h"
#include<iostream.h>
#include<assert.h>
#include"debug.h"

static EnvDebug debug("shadowsegment","shadowsegment");
#define DEBUGOUT cerr

SegmentVersion *SegmentVersion::init(char *svpuffer,int len)
{
  SegmentVersion *ret = (SegmentVersion *)svpuffer;
  ret->count_ = 0;
  Pageentry bpe;
  bpe.shadowbit=0;
  bpe.pageno=~0;

  for(unsigned int i=0;i< ret->maxcount();i++)
  {
    (*ret)[i]=bpe;
  }

  return ret;
}

int SegmentVersion::clearallshadowbits()
{
  int anz=0;
  for(unsigned int i=0;i<count();i++)
  {
    if((*this)[i].shadowbit)
    {
      (*this)[i].shadowbit=0;
      anz++;
    }
  }
  return anz;
}

void SegmentVersion::print(ostream& str)
{
  for(index_t i=0;i<count();i++)
  {
    str << i << ( entry[i].shadowbit ? "*" : " " );
    if(entry[i].pageno == UNDEFINED_PAGE) str << "UNDEF";
    else str << entry[i].pageno;
    str << endl;
  }
  str << endl;
}

/*------------ Shadow Master --------------*/
ShadowSegMaster::ShadowSegMaster()
    :tws(),rseg(&tws),
    mc(&rseg,((MasterLayout *)rseg.begin())->mc),
    m0(&rseg,((MasterLayout *)rseg.begin())->m0),
    m1(&rseg,((MasterLayout *)rseg.begin())->m1)
{
}

int ShadowSegMaster::open(char *name)
{
  debug.message(1,"ShadowSegMaster::open(%s)\n",name);
  if(rseg.open(name)==1)
  { 
    debug.message(1,"\tInitialisation ShadowSegMaster\n");
    m0.setoff(sizeof(MasterLayout));
    m0.ptr()->init(MAXPAGES);
    m1.setoff(sizeof(MasterLayout)+m0.ptr()->size());
    m1.ptr()->init(MAXPAGES);
    mc=m0;
    setstatus(0);
    rseg.write_page(0);
    return 1;
  }
  else
  {
    debug.message(1,"\tread ShadowSegMaster\n");
    rseg.read_page_reliable(0);
    return 0;
  }
}

void ShadowSegMaster::save()
{
  debug.message(1,"Saving Segmaster\n");
  rseg.save();
}

void ShadowSegMaster::copyandswitchm()
{
  RAMS::RPointer<BitMap> mn;
  debug.message(1,"copyandswitchm()\n");
  mn = (mc==m0 ? m1 : m0);
  memcpy(mn.ptr(),mc.ptr(),mc.ptr()->size());
  mc=mn;
}

void ShadowSegMaster::druckebelegtliste()
{
  DEBUGOUT << "Belegt in CurrentM:";
  ::druckebelegtliste(DEBUGOUT,CurrentM());
}

/*----------------- Shadowsegment -----------------*/
void ShadowSegment::init()
{
  debug.message(1,"Initialisation Segments\n");
  memset(svpuffer,'\0',2*PAGESIZE);
  vw[0]=SegmentVersion::init(svpuffer,PAGESIZE);
  vw[1]=SegmentVersion::init(svpuffer+PAGESIZE,PAGESIZE);
  savesegversion(0,0);
  savesegversion(1,1);
}

void ShadowSegment::clearshadowbitsandfree()
{
  debug.message(1,"ShadowSegment::clearshadowbitsandfree:\n");
  if(debug.level(3)) DEBUGOUT << "Vwork.count = " << Vwork().count() << endl;
  for(index_t i=0;i<Vwork().count();i++)
  {
    if(debug.level(3))
    {
      debug  << i << (Vwork()[i].shadowbit ? "*" : " ") 
	     << Vwork()[i].pageno
	     << endl;
    }
    if(Vwork()[i].shadowbit)
    {
      if(Vold()[i].pageno != UNDEFINED_PAGE)
      {
	if(debug.level(3)) DEBUGOUT << "\tFree old Page " << Vold()[i].pageno << endl;
	ssm.CurrentM().set(Vold()[i].pageno,false);
      }
      else
      {
	if(debug.level(3)) DEBUGOUT << "\tWas a new page " << endl;
      }
      Vwork()[i].shadowbit = 0;
    }
  }
}

void ShadowSegment::open(const char *name) throw (Systemerror)
{
  char mastername[64];
  char segname[64];
  char versionname[64];
  int r;

  debug.message(1,"ShadowSegment::open(%s)\n",name);
  snprintf(mastername,64,"%s.master",name);
  snprintf(versionname,64,"%s.version",name);
  snprintf(segname,64,"%s.seg",name);

  vfd = ::open(versionname,O_RDWR|O_CREAT,0660);
  if(vfd < 0)
  {
    throw Systemerror(errno,"open","ShadowSegment::init:open(%s):%m",versionname);
  }
  if(ssm.open(mastername)==1)
  {
    init();
  }
  else
  {
    if(ssm.getstatus()==1)
    {
       recover();
    }
    readsegversions();
    if(debug.level(2)) ssm.druckebelegtliste();
  }
  if(debug.level(2))
  {
    DEBUGOUT << "V0:" << endl; Vwork().print(DEBUGOUT);
    DEBUGOUT << "V1:" << endl; Vold().print(DEBUGOUT);
  }
  /** Gebe alte Seiten Frei und lsche Schadowbits - Abweichung **/ 
  clearshadowbitsandfree(); 
  if(debug.level(2))
  {
    DEBUGOUT << "Nach clearshadowbitsandfree():" << endl;
    DEBUGOUT << "V0:" << endl; Vwork().print(DEBUGOUT);
    DEBUGOUT << "V1:" << endl; Vold().print(DEBUGOUT);
  }
  /** Kopiere Vi0 nach Vi1 auf der Platte */
  savesegversion(0,1);
  if(debug.level(3))
  {
    DEBUGOUT << "Nach savesegversion(0,1):" << endl;
    DEBUGOUT << "V0:" << endl; Vwork().print(DEBUGOUT);
    DEBUGOUT << "V1:" << endl; Vold().print(DEBUGOUT);
  }
  /** Setze Status=1 */
  ssm.setstatus(1);
  /** schreibe Master auf Platte **/
  ssm.save();
  if(debug.level(2)) ssm.druckebelegtliste();
  /** loesche alle Shadowbits in vi0 */
  /// Abweichung vw[0]->clearallshadowbits();

  sfd = ::open(segname,O_RDWR|O_CREAT,0660);
  if(sfd < 0) 
  {
    throw Systemerror(errno,"open","ShadowSegment::init:open(%s):%m",segname);
  }

  char *alloc=NULL;
  char *first=NULL;
  size_t lang=0;
  if(debug.level(3))
  {
    DEBUGOUT << "Mappe " << Vwork().count() << " Seiten ..." << endl;
  }
  if(Vwork().count()==0)
  {
    // grow
    int anzseiten=10;
    debug.message(1,"Neues Segment, lege %d  Seiten an ...",anzseiten);
    Vwork().grow(anzseiten);
    long lang = anzseiten*getpagesize();
    int f0 = ::open("/dev/zero",O_RDONLY);
    if(r==-1) throw Systemerror(errno,"open","ShadowSegment::open(/dev/zero):%m");
    char *b=(char *)mmap(NULL,lang,PROT_READ,MAP_PRIVATE,f0,0);
    if(b==MAP_FAILED) 
    {
      throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
    } 
    set(b,lang);

    if(debug.level(1))
    {
      DEBUGOUT << "Segment gemappt: " << (void *)begin()
	    << " - " << (void *)end() << endl;
    }
    return; // 1;
  }
  {
    /* Reservierung, denn wer weiss wo die Seiten plaziert werden */
    int anzseiten=10;
    long lang = anzseiten*getpagesize();
    int f0 = ::open("/dev/zero",O_RDONLY);
    first=alloc=(char *)mmap(NULL,lang,PROT_READ,MAP_PRIVATE,f0,0);
    if(alloc==MAP_FAILED) 
    {
      throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
    } 
    munmap(alloc,lang);
    set(alloc,lang);
    if(debug.level(1))
    {
      DEBUGOUT << "Segment reserviert: " << (void *)begin()
	    << " - " << (void *)end() << endl;
    }
  }
  for(index_t i=0;i<Vwork().count();i++)
  {
    if(Vwork()[i].pageno != UNDEFINED_PAGE)
    {
      off_t off = (Vwork()[i].pageno  << 12);
      if(debug.level(3))
      {
	DEBUGOUT << "Seite " << i << " Block " << Vwork()[i].pageno 
	      << " " << (void *)alloc << " <- " << off << endl;
      }
      char *b=(char *)mmap(alloc,PAGESIZE,PROT_READ,MAP_SHARED,sfd,off);
      if(b==MAP_FAILED) 
      {
	throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
      }
      if(!alloc)
      {
	alloc = b;
	first=b;
	if(debug.level(3))
	{
	  DEBUGOUT << "Erstallocation bei " << (void *)alloc << endl;
	}
      }      
      else if(b != alloc)
      {
	throw Systemerror(0,"mmap","Shadowsegment::open: gewuenschte Speicheradresse nicht erhalten ");
      }
    }
    alloc+=PAGESIZE;
    lang +=PAGESIZE;
  }
  set(first,lang);
  if(debug.level(1))
  {
    DEBUGOUT << "Segment gemappt: " << (void *)begin()
	  << " - " << (void *)end() << endl;
  }
  // return; // 0;
}

void ShadowSegment::recover()
{
  debug.message(1,"Recover from Crash\n");
  readsegversions();
  savesegversion(1,0);
  if(debug.level(2))
  {
    DEBUGOUT << "V0 und V1:" << endl;Vwork().print(DEBUGOUT);
    ssm.druckebelegtliste();
  }
  debug.message(2,"\tSetze status=0\n");
  ssm.setstatus(0);
  ssm.save();
}  
  

void ShadowSegment::handle(void *page,void *addr)
{
  index_t dp;
  off_t dpo;
  index_t pi = offset2index((char *)page - (char *)begin()); 
  if(debug.level(1))
  {
    debug<< "Handle(" << page << ") = Page(" << pi << ")" << endl;
  }

  dp= ssm.CurrentM().find0();
  ssm.CurrentM().set(dp,true);

  if(debug.level(2))
  {
    DEBUGOUT << "alte Seite: ";
    if(Vwork()[pi].pageno == UNDEFINED_PAGE )
      DEBUGOUT << "undefined";
    else 
      DEBUGOUT <<  Vwork()[pi].pageno;
  }
  
  dpo= index2offset(dp);
  if(debug.level(2))
  {
    DEBUGOUT << " neue Seite: " << dp << " (Offset=" << dpo << ")" << endl;
  }

  int r=lseek(sfd,dpo,SEEK_SET);
  if(r<0) { perror("lseek"); exit(1); }
  r=write(sfd,page,PAGESIZE);
  if(r<0) { perror("write"); exit(1); }
  r=munmap(page,PAGESIZE);
  if(r<0) { perror("munmap"); exit(1); }
  /// schreibbar, denn in zukunft wird gleich der Block modifiziert
  void *newp = mmap(page,PAGESIZE,PROT_READ|PROT_WRITE,MAP_FIXED|MAP_SHARED,sfd,dpo);
  if(newp==(void *)(-1)) { perror("mmap"); exit(1); }

  Vwork()[pi].pageno=dp;
  Vwork()[pi].shadowbit=1;
}

void ShadowSegment::readsegversions() throw(Systemerror)
{
  debug.message(1,"readsegversions()\n");
  char *p=svpuffer;
  int rest=2*PAGESIZE;

  int nr=0;
  while(rest>0 && (nr= ::read(vfd,p,rest))>0)
  {
    p+=nr;
    rest-=nr;
  }
  if(nr < 0) throw Systemerror(errno,"read",
			       "ShadowSegment::readsegversions: read:%m");
  vw[0]=(SegmentVersion *)svpuffer;
  vw[1]=(SegmentVersion *)(svpuffer+PAGESIZE);
}

void ShadowSegment::savesegversion(int nr,int snr) throw(Systemerror)
{
  debug.message(1,"savesegversion v%d auf Platte v%d\n",nr,snr);
  assert(0<=nr && nr<2);
  assert(0<=snr && snr<2);
  off_t fo = snr*PAGESIZE;
  int r=lseek(vfd,fo,SEEK_SET);
  if(r<0) throw Systemerror(errno,"lseek","savesegversion(%d):lseek:%m",nr);
  r=write(vfd,(char *)vw[nr],PAGESIZE);
  if(r<0) throw Systemerror(errno,"write","savesegversion(%d):write:%m",nr);
  if(nr!=snr)
  {
    memcpy((char *)vw[snr],(char *)vw[nr],vw[nr]->size());
  }
}

void ShadowSegment::save()
{
  debug.message(1,"------- Save Segment --------\n");
  if(debug.level(2)) ssm.druckebelegtliste();
  ssm.copyandswitchm();
 
  if(debug.level(2)) ssm.druckebelegtliste();
  
  for(index_t i=0;i<Vwork().count() && i<<12 < length();i++)
  {
    off_t o  = index2offset(i);;
    off_t fo = index2offset(Vwork()[i].pageno);
   
    if(Vwork()[i].pageno == UNDEFINED_PAGE)
    {
      if(debug.level(3))
	DEBUGOUT << "Seite " << i << " unbenutzt " << endl;
    }
    else if(Vwork()[i].shadowbit)
    { 
      if(debug.level(3))
	DEBUGOUT << "Speichere Seite " << i << " (" << o << ") "
	      << "-> " << Vwork()[i].pageno << " (" << fo << ")" << endl;
      void *addr = (char *)begin()+o; 
      int r=msync(addr,PAGESIZE,MS_SYNC);
      if(r<0) { perror("msync"); exit(1); }
      r=mprotect(addr,PAGESIZE,PROT_READ);
      if(r<0) { perror("mprotect"); exit(1); }
#if 0 // With Write
      int r=lseek(sfd,fo,SEEK_SET);
      if(r<0) { perror("lseek"); exit(1); }
      r=write(sfd,akt.ptr(),PAGESIZE);
      if(r<0) { perror("write"); exit(1); }
#endif
    }
    else
    {
      if(debug.level(3)) DEBUGOUT << "Seite " << i  << " unveraendert " << endl;
    }
  }
  fsync(sfd);
  savesegversion(0,0);
  ssm.setstatus(0);
  if(debug.level(2)) ssm.druckebelegtliste();
  ssm.save();
  debug.message(1,"End von Save \n");
  if(debug.level(2))
  {
    DEBUGOUT << "V0:" << endl; Vwork().print(DEBUGOUT);
    DEBUGOUT << "V1:" << endl; Vold().print(DEBUGOUT);
  }
}


void ShadowSegment::reopen()
{
  clearshadowbitsandfree();
  savesegversion(0,1);
  ssm.setstatus(1);  
  ssm.save();
}

void ShadowSegment::rollback()
{
  debug.message(1,"rollback\n");
  for(index_t i=0;i<Vwork().count() && i<<12 < length();i++)
  {
    if(Vwork()[i].shadowbit)
    {
      if(debug.level(3))
      {
	DEBUGOUT << i << " instead " << Vwork()[i].pageno
		 << " -> " << Vold()[i].pageno << endl;
      }
     
  
      // Seite mu wieder Richtig gemappt werden
      off_t o=i<<12;
      off_t old_off=Vold()[i].pageno <<12;
      void *addr = (char *)begin()+o;
      if(munmap(addr,PAGESIZE)<0)
      {
	perror("munmap");exit(1);
      }
      ssm.CurrentM().set(Vwork()[i].pageno,false);
      char *b=(char *)mmap(addr,PAGESIZE,PROT_READ,MAP_SHARED,sfd,old_off);
      assert(b!=MAP_FAILED && (void *)b==addr);
      Vwork()[i] = Vold()[i];
    }
  }
}









